#############################################################################
# tmap.tcl Copyright 2004, 2005, 2006 r0lZ
# VTS_TMAPTI table stuff.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::tmap {


# Rebuild the whole or a part of the VTS_TMAPTI table of the VTS.
# NOTE: This function retrieves the nav packs in the VOB files and can be very slow!
# The cell lists of the PGCs must be coherent before calling this function!
#
# If onlypgc == 0 (the default), the whole table is recomputed.
# If onlypgc > 0, only this PGC (and the missing PGCs in the table) are recomputed.
# If onlypgc < 0, only the missing PGCs in the table are recomputed.
#
# Example usage: ::tmap::rebuild_tmap_from_vobs 1 1
proc rebuild_tmap_from_vobs {vts {onlypgc 0}} {
	set numpgcs $::pgcs($vts,numpgcs)

	# display error message if VOB is desynchronized
	unset -nocomplain ::menued::load_nav_pack_error_skip

	# build list of time maps to process
	if {$::raw_sectors($vts,VTS_TMAPTI)==""} {
		set existingtimemaps 0
	} else {
		set existingtimemaps [get_int ::raw_sectors($vts,VTS_TMAPTI) 0 2]
	}
	set pgclist {}
	for {set pgc 1} {$pgc<=$::pgcs($vts,numpgcs)} {incr pgc} {
		if {$onlypgc == 0 || $onlypgc == $pgc || $pgc > $existingtimemaps} {
			lappend pgclist $pgc
		}
	}

	# compute total number of entries to process (for progress bar)
	set maxentries 0
	foreach pgc $pgclist {
		if {$::pgcs($vts,$pgc,numcells)==0} continue
		set pgcsecs [expr {[BCD2Secs [string range $::pgcs($vts,$pgc,header) 4 7]]}]
		if {$pgcsecs<1} continue
		set timeunit [expr {($pgcsecs+2047)/2048}]
		incr maxentries [expr {$pgcsecs/$timeunit}]
	}
	set refreshevery [expr $maxentries/50]
	if {$refreshevery < 1} {
		set refreshevery 1
	}
	set thisentry 0
	set oldvobfile "*:*"
	set fh ""

	# for each PGC...
	for {set pgc 1} {$pgc<=$::pgcs($vts,numpgcs)} {incr pgc} {
		# skip tmap if it is not needed/requested
		if {[lsearch $pgclist $pgc]==-1} {
			set tmapoffset [get_int ::raw_sectors($vts,VTS_TMAPTI) [expr {4+($pgc*4)}] 4]
			set tmap($pgc,TimeUnit) [get_int ::raw_sectors($vts,VTS_TMAPTI) $tmapoffset 1]
			set tmap($pgc,NumEntries) [get_int ::raw_sectors($vts,VTS_TMAPTI) [expr {$tmapoffset+2}] 2]
			if {$tmap($pgc,NumEntries)==0} {
				set tmap($pgc) ""
			} else {
				set tmap($pgc) [string range $::raw_sectors($vts,VTS_TMAPTI) [expr {$tmapoffset+4}] \
						[expr {$tmapoffset+3+($tmap($pgc,NumEntries)*4)}]]
			}
			continue
		}

		set numcells $::pgcs($vts,$pgc,numcells)
		# empty tmap for a dummy PGC
		if {$numcells==0} {
			set tmap($pgc,TimeUnit) 0
			set tmap($pgc,NumEntries) 0
			set tmap($pgc) ""
			continue
		}

		# Compute time unit for this PGC
		set cellpbtable $::pgcs($vts,$pgc,cellpbtable)
		set pgcsecs [expr {[BCD2Secs [string range $::pgcs($vts,$pgc,header) 4 7]]}]
		if {$pgcsecs<1} {
			set tmap($pgc,TimeUnit) 0
			set tmap($pgc,NumEntries) 0
			set tmap($pgc) ""
			continue
		}
		set timeunit [expr {($pgcsecs+2047)/2048}]
		set timeunitticks [expr {$timeunit*90000}]

		set curpgctime $timeunitticks		;# timing used to select or reject the current nav pack
		set curcellstarttime 0

		set tmap($pgc,TimeUnit) $timeunit
		set tmap($pgc) ""

		set celltimeoffset 0
		for {set cell 1} {$cell<=$numcells} {incr cell} {
			set cellpbidx [expr {($cell-1)*24}]

			# skip angle cells when angle# > 1
			set angletype [expr {[get_int cellpbtable $cellpbidx 1]/16}]
			if {$angletype>5} continue

			show_msg "Building time map for VTS $vts, PGC $pgc (Cell $cell) ..." $thisentry $maxentries
			set firstVOBU [get_int cellpbtable [expr {$cellpbidx+8}] 4]
			set firstILVU [get_int cellpbtable [expr {$cellpbidx+12}] 4]
			set lastVOBU  [get_int cellpbtable [expr {$cellpbidx+16}] 4]
			set lba $firstVOBU

			set poslist [::menued::get_nav_pack_position pgc $vts $lba $::dvddir]
			set vobfile [lindex $poslist 0]
			set sector  [lindex $poslist 1]
			if {$vobfile==""} {
				if {$fh!=""} {close $fh}
				vob_warning $vts $pgc $cell "Missing [file tail $vobfile]"
				return -1
			}
			if {$oldvobfile != $vobfile} {
				if {$fh!=""} {close $fh}
				set fh [::menued::openbinfile $vobfile r]
				if {$fh==""} {
					vob_warning $vts $pgc $cell "Can't open [file tail $vobfile]"
					return -2
				}
				set oldvobfile $vobfile
			}
			set navpack [::menued::load_nav_pack $vobfile $fh $sector]
			if {$navpack==""} {
				if {$fh!=""} {close $fh}
				vob_warning $vts $pgc $cell \
						"First Nav Pack in [file tail $vobfile] is not valid"
				return -3
			}
			set nextlba $lba
			set vid [get_int navpack 1055 2]
			set cid [get_int navpack 1058 1]
	
			# compute new time offset for current cell
			set celltimeoffset [expr {$celltimeoffset-[get_int navpack 57 4]}]

			while {$nextlba <= $lastVOBU} {
				set lba $nextlba
				set poslist [::menued::get_nav_pack_position pgc $vts $lba $::dvddir]
				set vobfile [lindex $poslist 0]
				set sector  [lindex $poslist 1]
				if {$vobfile==""} {
					if {$fh!=""} {close $fh}
					vob_warning $vts $pgc $cell "Missing [file tail $vobfile]"
					return -4
				}
				if {$vobfile!=$oldvobfile} {
					if {$fh!=""} {close $fh}
					set fh [::menued::openbinfile $vobfile r]
					if {$fh==""} {
						vob_warning $vts $pgc $cell "Can't open [file tail $vobfile]"
						return -5
					}
					set oldvobfile $vobfile
				}
				set navpack [::menued::load_nav_pack $vobfile $fh $sector]
				if {$navpack==""} {
					if {$fh!=""} {close $fh}
					vob_warning $vts $pgc $cell \
							"Nav Pack at LBA $lba in [file tail $vobfile] is not valid"
					return -6
				}
				set nextlba [expr {$lba+[get_int navpack 1039 4]+1}]
				if {$firstILVU>0 && ($vid!=[get_int navpack 1055 2] || $cid!=[get_int navpack 1058 1])} {
					continue
				}
				set endnavtime [expr {[get_int navpack 61 4]+$celltimeoffset}]
				#while {[expr {[get_int navpack 57 4]+$celltimeoffset}]<=$curpgctime && $curpgctime<$endnavtime} 
				while {$curpgctime<$endnavtime} {
					set tmp \000\000\000\000
					set_int tmp 0 $lba 4
					set tmap($pgc) $tmap($pgc)$tmp
					incr curpgctime $timeunitticks
					incr thisentry
					if {[expr {$thisentry%$refreshevery}]==0} {
						update_progress $thisentry $maxentries
					}
				}
			}
			set celltimeoffset $endnavtime
			if {$tmap($pgc)!=""} {
				set idx [expr {[string length $tmap($pgc)]-4}]
				set_int tmap($pgc) $idx [expr {[get_int tmap($pgc) $idx 4]|0x80000000}] 4
			}
		}
		set tmap($pgc,NumEntries) [expr {[string length $tmap($pgc)]/4}]
		if {$tmap($pgc,NumEntries)==0} {
			set tmap($pgc,TimeUnit) 0
		}
	}
	update_progress 100 100

	# concat the parts of the table
	show_msg "Replacing VTS_TMAPTI of VTS $vts ..." 100 100
	set table \000\000\000\000\000\000\000\000
	set_int table 0 $numpgcs 2
	for {set pgc 1} {$pgc<=$numpgcs} {incr pgc} {
		set table $table\000\000\000\000
	}
	set idx 8
	for {set pgc 1} {$pgc<=$numpgcs} {incr pgc} {
		set_int table $idx [string length $table] 4
		set table $table\000\000\000\000
		set_int table [expr {[string length $table]-4}] $tmap($pgc,TimeUnit) 1
		set_int table [expr {[string length $table]-2}] $tmap($pgc,NumEntries) 2
		set table $table$tmap($pgc)
		incr idx 4
	}
	set_int table 4 [expr {[string length $table]-1}] 4
	
	set ::raw_sectors($vts,VTS_TMAPTI) $table
	show_msg "VTS_TMAPTI table of VTS $vts created successfully."
	::main::dvd_modified true false
	if {$fh!=""} {close $fh}
	return 0
}


# return a BCD time in seconds (Frames are ignored)
proc BCD2Secs {bcd} {
	set tmp [get_int bcd 0 1]
	set secs [expr {[format %i [expr {$tmp/16}]]*36000}]
	set secs [expr {$secs+[format %i [expr {$tmp&15}]]*3600}]
	set tmp [get_int bcd 1 1]
	set secs [expr {$secs+[format %i [expr {$tmp/16}]]*600}]
	set secs [expr {$secs+[format %i [expr {$tmp&15}]]*60}]
	set tmp [get_int bcd 2 1]
	set secs [expr {$secs+[format %i [expr {$tmp/16}]]*10}]
	return [expr {$secs+[format %i [expr {$tmp&15}]]}]
}

proc vob_warning {vts pgc cell msg} {
	set msg "WARNING:\nError creating time map for VTS $vts, PGC $pgc, cell $cell:\n$msg !"
	set msg "$msg\nVTS_TMAPTI table for VTS $vts NOT modified!"
	show_warning -message $msg -title "Build VTS_TMAPTI table" -icon warning -parent .
}

proc rebuild_tmap_cur_vts {} {
	set vts [lindex [split [lindex $::editor_cmds_from 1] ,] 0]
	if {$vts == 0} {
		bell
		return
	}
	rebuild_tmap_from_vobs $vts
	dont_need_new_tmap $vts
}

proc rebuild_tmap_cur_pgc {} {
	if {[lindex $::editor_cmds_from 0]=="menu"} {
		bell
		return
	}
	set idxs [split [lindex $::editor_cmds_from 1] ,]
	set vts [lindex $idxs 0]
	if {$vts == 0} {
		bell
		return
	}
	set pgc [lindex $idxs 1]
	rebuild_tmap_from_vobs $vts $pgc
	dont_need_new_tmap $vts $pgc
}

proc rebuild_all_tmaps {} {
	set errormsg ""
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set rc [rebuild_tmap_from_vobs $vts]
		if {$rc != 0} {
			set errormsg "$errormsg\nError $rc creating VTS_TMAPTI table for VTS $vts !"
		}
	}
	if {$errormsg == ""} {
		show_msg "Time maps regenerated successfully for all titlesets."
	} else {
		show_msg "Error rebuilding time maps for some titlesets!"
		tk_messageBox -title "Rebuild all VTS_TMAPTI tables of the DVD" -icon warning -parent . \
			-message "$errormsg\n\nThe VOB files must be present and valid to use this function!"
	}
	set ::tmap::need_new_tmap_list {}
	return
}

# adds a PGC or VTS in the global list holding the PGCs to be processed when the DVD is saved.
proc need_new_tmap {vts {onlypgc 0}} {
	if {$vts == 0} return
	set entry [list $vts $onlypgc]
	if {[info exists ::tmap::need_new_tmap_list]} {
		dont_need_new_tmap $vts $onlypgc
	} else {
		set ::tmap::need_new_tmap_list {}
	}
	lappend ::tmap::need_new_tmap_list $entry
	return
}

# adds a PGC or VTS in the global list holding the PGCs to be processed when the DVD is saved.
proc dont_need_new_tmap {vts {onlypgc 0}} {
	set entry [list $vts $onlypgc]
	if {[info exists ::tmap::need_new_tmap_list]} {
		if {$onlypgc == 0} {
			for {set i 0} {$i < [llength $::tmap::need_new_tmap_list]} {incr i} {
				if {[lindex [lindex $::tmap::need_new_tmap_list $i] 0] == $vts} {
					set ::tmap::need_new_tmap_list [lreplace $::tmap::need_new_tmap_list $i $i]
				}
			}
		} else {
			set i [lsearch $::tmap::need_new_tmap_list $entry]
			if {$i >= 0} {
				set ::tmap::need_new_tmap_list [lreplace $::tmap::need_new_tmap_list $i $i]
			}
		}
	}
	return
}

proc rebuild_needed_tmaps {} {
	if {! [info exists ::tmap::need_new_tmap_list]} return

	set lst [lsort -dictionary $::tmap::need_new_tmap_list]
	foreach i $lst {
		rebuild_tmap_from_vobs [lindex $i 0] [lindex $i 1]
	}
	set ::tmap::need_new_tmap_list {}
	return
}

} ;# end of ::tmap namespace
